==> spellathon-5.59.tar.gz: spellathon-5.59/src/logic/game.cc <==
/*
 * Copyright (c) 2009, Girish Rao
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * The names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT i
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#include <algorithm>

#ifndef SPTH_CL
#include <QCoreApplication>
#endif /* SPTH_CL */

#include "game.h"
#include "dict.h"
#include "util.h"

#ifdef WIN32
/* mingw doesn't have random(3) */
# define random rand
# define srandom srand
#endif /* WIN32 */

namespace Spellathon
{
	static int sidxcmp(const void *a, const void *b);

	Game::Game(Dict *dct, bool plural_sieve) 
		: dict_(dct), lexSort_(false), 
		  pluralSieve_(plural_sieve), score_(0), rating_("")
	{
		srandom(time(0));
		words_.reserve(128);
		wordPtrs_.reserve(128); 
	}

	void 
	Game::setPluralSieve(bool e)
	{ 
		pluralSieve_ = e;
	}

	bool
	Game::play(bool precomp, int minsol, int maxsol)
	{
		lexSort_ = false;
		if (precomp) 
			return _play_precomp(minsol, maxsol);
		else
			return _play(minsol, maxsol);
	}

	bool
	Game::_play(int minsol, int maxsol)
	{
		char			sig[MAX_LETRS+1];
		unsigned int 	(*sidx)[2];
		enum			{IDX, RAND};

	    sidx = (unsigned int (*)[2])malloc(2 * dict_->nent7() * sizeof(int));
		if (sidx == NULL) 
			spth_error(true, "Out of memory.");

		/* shuffle the indexes. Knuth shuffle? */
		int j = dict_->off7();
		if (j < 0)
			goto notfound;

        for (int i = 0; i < dict_->nent7(); i++) {
                sidx[i][IDX] = j++;
                sidx[i][RAND] = random();
        }
	
        qsort((void*)sidx, dict_->nent7(), 2 * sizeof(int), sidxcmp);
	
		unsigned int cidx[MAX_LETRS][2];
        for (int i = 0; i < MAX_LETRS; i++) {
                cidx[i][IDX] = i;
                cidx[i][RAND] = random();
        }
	
        qsort((void*)cidx, MAX_LETRS, 2 * sizeof(int), sidxcmp);
	
		/*
		 * find a random seven letter signature from the dict and solve
		 * it and any of the six different central letter combinations.
		 * consider the word playable if MIN_SOL <= solutions <= MAX_SOL
		 *
		 * this technique works well. it's fast, sufficiently
		 * random and covers all the signatures.
		 */
		for (int i = 0; i < dict_->nent7(); i++) {

			const char *dsig = (*dict_).dictarr[sidx[i][IDX]].sign;

			int j = 0;
			while (j < MAX_LETRS) {
				sig[j] = dsig[cidx[j][IDX]]; 
				j++;
			}
			sig[j] = 0;

			int k = MAX_LETRS - 2;
			do {
				if (solver(sig, minsol, maxsol) != -1) {
					/* found suitable sign/game */
					free(sidx);
					puzzle_ = sig;
					return true;
				}

				if (k < 0)
					break;

				/* try a different central letter */	
				swapch(&sig[MAX_LETRS-1], &sig[k--]);
	
			} while (1);

			/* 
			 * this can take some time in pathological cases 
			 * so we need to keep the gui responsive 
			 */	
			if (!(i & ((1<<8) - 1)) && i > 1024) 
#ifdef SPTH_CL
				spinner();
#else
				QCoreApplication::processEvents();
#endif /* SPTH_CL */
		}
notfound:
		/* Uh Oh */	
		puzzle_ = "";
		free(sidx);
		return false;
	}

	bool
	Game::_play_precomp(int minsol, int maxsol)
	{
		int i = 0, s = -1, e = -1;

		while (i < dict_->precomp_nent()) {
			if (dict_->precomparr[i].nsol >= minsol) {
				s = i;
				break;
			}
			i++;
		}

		while (i < dict_->precomp_nent()) {
			if (dict_->precomparr[i].nsol > maxsol) {
				break;
			}
			i++;
		}

		e = i;

		if (s < 0 || s >= e) {
			puzzle_ = "";
			return false;
		} else {
			int r = randnum(s, e);
			puzzle_ = dict_->precomparr[r].sign;
			solver(puzzle_.c_str());
			return true;
		}
	}

	bool
	Game::solve(string letters)
	{
		lexSort_ = true;
		solver(letters.c_str());
		return nWords() > 0;
	}

	static int 
	sidxcmp(const void *a, const void *b)
	{
	        const unsigned int (*c)[2] = (const unsigned int (*)[2])a;
	        const unsigned int (*d)[2] = (const unsigned int (*)[2])b;
	
	        if ((*c)[1] == (*d)[1])
	                return 0;
	        if ((*c)[1] < (*d)[1])
	                return -1;
	        else
	                return 1;
	}

	/* performance critical routine */
	int
	Game::solver(const char* letrs, int min, int max)
	{
		char	signarr[NCOMBINATIONS][MAX_LETRS+1];
	
		int c = signGen(letrs, signarr);
	
		wordPtrs_.clear();
		/* 
		 * for all the sign. generated by signgen() find matching 
		 * sign (and thus words) from the dictionary
		 */
		if (pluralSieve_) {
			/* remove plurals */
			bool has7 = false;
			vector<char *> w;
			w.reserve(64);

			for (int k = 0; k < c; k++) {
				w.clear();
				dict_->matchSign(signarr[k], w);
				int i = (int)w.size();

				while (--i >= 0) {
					char word[MAX_LETRS+1];
					strcpy7(word, w[i]);
					int slen_orig = strlen(word);
					int slen = slen_orig;
					/* .*s */
					if (word[--slen] == 's') {
						word[slen] = 0;	
						if (dict_->matchWord(word)) 
							continue;
						/* .*es */
						if (word[--slen] == 'e') {
							word[slen] = 0;
							if (dict_->matchWord(word)) 
								continue;
						}
						/* .*ies */
						if (word[--slen] == 'i') {
							word[slen] = 'y';
							if (dict_->matchWord(word)) 
								continue;
						}
					}
					/* word accepted */
					has7 = (slen_orig == 7);
					wordPtrs_.push_back(w[i]); 
				}

				if ((int)wordPtrs_.size() > max)
					return -1;
			}

			if (!has7) 
				return -1;
		} else { 
			for (int i = 0; i < c; i++) {
				dict_->matchSign(signarr[i], wordPtrs_);
				if ((int)wordPtrs_.size() > max)
					return -1;
			}
		}
	
		if ((int)wordPtrs_.size() < min)
			return -1;

		return wordPtrs_.size();
	}
	
	/* 
	 * find all the letter combinations of length 4, 5, 6 and 7.
	 * the letters can be duplicates so that needs to be handled.
	 */
	/* performance critical routine */
	int
	Game::signGen(const char *letters, char signarr[][MAX_LETRS+1])
	{
		int		i, j, k;
		const uint8_t	smask[NCOMBINATIONS] = { 
			/* C(6,3) */
			0x7,  0xb,  0xd,  0xe,  0x13, 0x15, 
			0x16, 0x19, 0x1a, 0x1c, 0x23, 0x25, 
			0x26, 0x29, 0x2a, 0x2c, 0x31, 0x32, 
			0x34, 0x38,
			/* C(6,4) */
			0xf,  0x17, 0x1b, 0x1d, 0x1e, 0x27, 
			0x2b, 0x2d, 0x2e, 0x33, 0x35, 0x36, 
			0x39, 0x3a, 0x3c,
			/* C(6,5) */
			0x1f, 0x2f, 0x37, 0x3b, 0x3d, 0x3e,
			/* C(6,6) */
			0x3f 
		}; 
		int8_t coff[] = { -1, -1, -1, -1, -1, -1, -1, -1 };
	
		memset(signarr, 0, NCOMBINATIONS * (MAX_LETRS+1));

		int n = 0;

		char psortedletrs[MAX_LETRS+1];
		strcpy7(psortedletrs, letters);

		sign(psortedletrs, MAX_LETRS-1);

		char letrlookup[64];
		letrlookup[0] = 0;
		letrlookup[0x1] = psortedletrs[0];
		letrlookup[0x2] = psortedletrs[1];
		letrlookup[0x4] = psortedletrs[2];
		letrlookup[0x8] = psortedletrs[3];
		letrlookup[0x10] = psortedletrs[4];
		letrlookup[0x20] = psortedletrs[5];

		/* 
		 * the central letter is present in all the words so
		 * find the combinations of the remaining 6 letters 
		 */
		for (i = 0; i < NCOMBINATIONS; i++) {
			j = 0;
			uint8_t maskval = smask[i];
			char *signtmp = &signarr[n][0];

			/* generate a letter combination from the bitmask */
			signtmp[j] = letrlookup[maskval&1	]; j +=  maskval&1;
			signtmp[j] = letrlookup[maskval&0x2	]; j += (maskval&0x2) >>1;
			signtmp[j] = letrlookup[maskval&0x4	]; j += (maskval&0x4) >>2;
			signtmp[j] = letrlookup[maskval&0x8	]; j += (maskval&0x8) >>3;
			signtmp[j] = letrlookup[maskval&0x10]; j += (maskval&0x10)>>4;
			signtmp[j] = letrlookup[maskval&0x20]; j += (maskval&0x20)>>5;
	
			/* add the central letter in the correct sorted position */
			int p = j;
			int centraletr = psortedletrs[MAX_LETRS-1];
			while (p > 0 && centraletr < signtmp[p-1]) {
				signtmp[p] = signtmp[p-1];
				--p;
			}
			signtmp[p] = centraletr;
	
			if (coff[j] == -1)
			   coff[j] = n;

			/* ignore duplicate comb. in the case where letters 
			 * are duplicates */
			k = n-1;
			while (k >= coff[j]) {
				if (strcmp7_eq(signarr[n], signarr[k])) {
					n--;	
					break;
				}
				k--;
			}
			n++;
		}
	
		return n;	
	}

	const vector<string> & 
	Game::words() 
	{ 
		words_.clear();

		for (int i = 0; i < (int)wordPtrs_.size(); i++) 
			words_.push_back(wordPtrs_[i]); 

		if (lexSort_)
			std::sort(words_.begin(), words_.end());

		return words_;
	}
	void
	Game::calcScore(int nw, int nt, int h, int t, bool entered7)
	{
		double s = 100;
	
		/* time penalty */
		if ((nt <= 5 && t/nt > 10)  ||
		    (nt <= 10 && t/nt > 15) ||
		    (nt <= 20 && t/nt > 20) ||
		     t/nt > 30)
			s -= TimePenalty;
	
		/* hint penalty */
		switch (h) {
		case 1:
			s -= HintPenalty[0];
			break;
		case 2:
			s -= HintPenalty[1];
			break;
		}
	
		if (!entered7)
			s -= No7Penalty;

		/* fraction of total words correct */
		s *= nw/(double)nt;
		score_ = (int)s;
	
		if (score_ >= Scores[3]) {
			rating_ = Ratings[3];
			rating_ += '!';
			return;
		}
		if (score_ >= Scores[2]) {
			rating_ = Ratings[2];
			return;
		}
		if (score_ > Scores[1]) {
			rating_ = Ratings[1];
			return;
		}
		rating_ = Ratings[0];
	}
}

==> spellathon-5.72.tar.gz: spellathon-5.72/src/logic/game.cc <==
/*
 * Copyright (c) 2009, Girish Rao
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * The names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT i
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#include <stdint.h>
#include <cassert>
#include <algorithm>

#ifndef SPTH_CL
#include <QCoreApplication>
#endif /* SPTH_CL */

#include "game.h"
#include "dict.h"
#include "util.h"

#ifdef WIN32
/* mingw doesn't have random(3) */
# define random rand
# define srandom srand
#endif /* WIN32 */

namespace Spellathon
{
	static int sidxcmp(const void *a, const void *b);

	Game::Game(Dict *dct, bool plural_sieve) 
		: dict_(dct), lexSort_(false), 
		  pluralSieve_(plural_sieve), score_(0), rating_("")
	{
		srandom(time(0));
		words_.reserve(128);
	}

	void 
	Game::setPluralSieve(bool e)
	{ 
		pluralSieve_ = e;
	}

	bool
	Game::play(bool precomp, int minsol, int maxsol)
	{
		lexSort_ = false;
		if (precomp) 
			return _play_precomp(minsol, maxsol);
		else
			return _play(minsol, maxsol);
	}

	bool
	Game::_play(int minsol, int maxsol)
	{
		char			sig[MAX_LETRS+1];
		unsigned int 	(*sidx)[2];
		enum			{IDX, RAND};

		const DictEnt* da = dict_->dictArr();

	    sidx = (unsigned int (*)[2])malloc(2 * dict_->nent7() * sizeof(int));
		if (sidx == NULL) 
			spth_error(true, "Out of memory.");

		/* shuffle the dict. indexes. Knuth shuffle? */
		int j = dict_->off7();
		if (j < 0)
			goto notfound;

        for (int i = 0; i < dict_->nent7(); i++) {
                sidx[i][IDX] = j++;
                sidx[i][RAND] = random();
        }
	
        qsort((void*)sidx, dict_->nent7(), 2 * sizeof(int), sidxcmp);
	
		/* shuffled letter indexes */
		unsigned int cidx[MAX_LETRS][2];
        for (int i = 0; i < MAX_LETRS; i++) {
                cidx[i][IDX] = i;
                cidx[i][RAND] = random();
        }
	
        qsort((void*)cidx, MAX_LETRS, 2 * sizeof(int), sidxcmp);
	
		/*
		 * find a random seven letter signature from the dict and solve
		 * it and any of the six different central letter combinations.
		 * consider the word playable if MIN_SOL <= #solutions <= MAX_SOL
		 *
		 * this technique works well. it's fast, sufficiently
		 * random and covers all the signatures (all possible games).
		 */
		for (int i = 0; i < dict_->nent7(); i++) {

			/* get a random sign. from the dict. */
			const char *dsig = da[sidx[i][IDX]].sign;

			/* randomize the letters so that we don't the same central 
			   letter (& thus the same game) everytime this sign. is chosen */
			int j = 0;
			while (j < MAX_LETRS) {
				sig[j] = dsig[cidx[j][IDX]]; 
				j++;
			}
			sig[j] = 0;

			char cltried[MAX_LETRS];
			int k = MAX_LETRS - 2;
			int clidx = 0;
			memset(cltried, 0, sizeof(cltried));
			do {
				if (solver(sig, minsol, maxsol) != -1) {
					/* found a suitable game */
					free(sidx);
					puzzle_ = sig;
					return true;
				}

				/* find non-duplicate central letter */
				cltried[clidx++] = sig[MAX_LETRS-1];
again:
				for (char *p = cltried; *p && k >= 0; p++) {
					if (*p == sig[k]) {
						k--;
						goto again;
					} 
				}

				if (k < 0)
					break;

				/* solve for a different central letter */	
				swapch(&sig[MAX_LETRS-1], &sig[k--]);
	
			} while (1);

			/* 
			 * this can take some time in pathological cases 
			 * so we need to keep the gui responsive 
			 */	
			if (!(i & ((1<<9) - 1)) && i > 1024) 
#ifdef SPTH_CL
				spinner();
#else
				QCoreApplication::processEvents();
#endif /* SPTH_CL */
		}
notfound:
		/* Uh Oh */	
		puzzle_ = "";
		free(sidx);
		return false;
	}

	bool
	Game::_play_precomp(int minsol, int maxsol)
	{
		int i = 0, s = -1, e = -1;

		const PrecompEnt* pc = dict_->precompArr();

		while (i < dict_->precompSize()) {
			if (pc[i].nsol >= minsol) {
				s = i;
				break;
			}
			i++;
		}

		while (i < dict_->precompSize()) {
			if (pc[i].nsol > maxsol) {
				break;
			}
			i++;
		}

		e = i;

		if (s < 0 || s >= e) {
			puzzle_ = "";
			return false;
		} else {
			int r = randnum(s, e);
			puzzle_ = pc[r].sign;
			solver(puzzle_.c_str());
			return true;
		}
	}

	bool
	Game::solve(string letters)
	{
		lexSort_ = true;
		return solver(letters.c_str()) > 0;
	}

	static int 
	sidxcmp(const void *a, const void *b)
	{
	        const unsigned int (*c)[2] = (const unsigned int (*)[2])a;
	        const unsigned int (*d)[2] = (const unsigned int (*)[2])b;
	
	        if ((*c)[1] == (*d)[1])
	                return 0;
	        if ((*c)[1] < (*d)[1])
	                return -1;
	        else
	                return 1;
	}

	/* performance critical routine */
	int
	Game::solver(const char* letrs, int min, int max)
	{
		char	signarr[NCOMBINATIONS][MAX_LETRS+1];
	
		nWords_ = 0;
		int has7 = 0;
		int s = signGen(letrs, signarr);
	
		assert(min <= max);
		assert(max <= SOLVER_MAX_WORDS);

		for (int i = 0; i < s; i++) {
			dict_->matchSign(signarr[i], wordPtrs_, &nWords_, &has7);
			if (nWords_ > max)
				return -1;
		}
	
		if (nWords_ < min)
			return -1;

		if (!has7)
			return -1;

		return nWords_;
	}
	
	/* 
	 * find all the letter combinations of length 4, 5, 6 and 7.
	 * the letters can be duplicates so that needs to be handled.
	 */
	/* performance critical routine */
	int
	Game::signGen(const char *letters, char signarr[][MAX_LETRS+1])
	{
		int		i, j, k;
		const uint8_t	smask[NCOMBINATIONS] = { 
			/* C(6,3) */
			0x7,  0xb,  0xd,  0xe,  0x13, 0x15, 
			0x16, 0x19, 0x1a, 0x1c, 0x23, 0x25, 
			0x26, 0x29, 0x2a, 0x2c, 0x31, 0x32, 
			0x34, 0x38,
			/* C(6,4) */
			0xf,  0x17, 0x1b, 0x1d, 0x1e, 0x27, 
			0x2b, 0x2d, 0x2e, 0x33, 0x35, 0x36, 
			0x39, 0x3a, 0x3c,
			/* C(6,5) */
			0x1f, 0x2f, 0x37, 0x3b, 0x3d, 0x3e,
			/* C(6,6) */
			0x3f 
		}; 
		int8_t coff[] = { -1, -1, -1, -1, -1, -1, -1, -1 };
	
		memset(signarr, 0, NCOMBINATIONS * (MAX_LETRS+1));

		int n = 0;

		char psortedletrs[MAX_LETRS+1];
		strcpy7(psortedletrs, letters);

		sign(psortedletrs, MAX_LETRS-1);

		char letrlookup[64];
		letrlookup[0] = 0;
		letrlookup[0x1] = psortedletrs[0];
		letrlookup[0x2] = psortedletrs[1];
		letrlookup[0x4] = psortedletrs[2];
		letrlookup[0x8] = psortedletrs[3];
		letrlookup[0x10] = psortedletrs[4];
		letrlookup[0x20] = psortedletrs[5];

		/* 
		 * the central letter is present in all the words so
		 * find the combinations of the remaining 6 letters 
		 */
		for (i = 0; i < NCOMBINATIONS; i++) {
			j = 0;
			uint8_t maskval = smask[i];
			char *signtmp = &signarr[n][0];

			/* generate a letter combination from the bitmask */
			signtmp[j] = letrlookup[maskval&1	]; j +=  maskval&1;
			signtmp[j] = letrlookup[maskval&0x2	]; j += (maskval&0x2) >>1;
			signtmp[j] = letrlookup[maskval&0x4	]; j += (maskval&0x4) >>2;
			signtmp[j] = letrlookup[maskval&0x8	]; j += (maskval&0x8) >>3;
			signtmp[j] = letrlookup[maskval&0x10]; j += (maskval&0x10)>>4;
			signtmp[j] = letrlookup[maskval&0x20]; j += (maskval&0x20)>>5;
	
			/* add the central letter in the correct sorted position */
			int p = j;
			int centraletr = psortedletrs[MAX_LETRS-1];
			while (p > 0 && centraletr < signtmp[p-1]) {
				signtmp[p] = signtmp[p-1];
				--p;
			}
			signtmp[p] = centraletr;
	
			if (coff[j] == -1)
			   coff[j] = n;

			/* ignore duplicate comb. in the case where letters 
			 * are duplicates */
			k = n-1;
			while (k >= coff[j]) {
				if (strcmp7_eq(signarr[n], signarr[k])) {
					n--;	
					break;
				}
				k--;
			}
			n++;
		}
	
		return n;	
	}

	const vector<string> & 
	Game::words() 
	{ 
		words_.clear();
		for (int i = 0; i < nWords_; i++) 
			words_.push_back(wordPtrs_[i]); 

		if (lexSort_)
			std::sort(words_.begin(), words_.end());

		return words_;
	}

	void
	Game::calcScore(int nw, int nt, int h, int t, bool entered7)
	{
		double s = 100;
	
		/* time penalty */
		if ((nt <= 5 && t/nt > 10)  ||
		    (nt <= 10 && t/nt > 15) ||
		    (nt <= 20 && t/nt > 20) ||
		     t/nt > 30)
			s -= TimePenalty;
	
		/* hint penalty */
		switch (h) {
		case 1:
			s -= HintPenalty[0];
			break;
		case 2:
			s -= HintPenalty[1];
			break;
		}
	
		if (!entered7)
			s -= No7Penalty;

		/* fraction of total words correct */
		s *= nw/(double)nt;
		score_ = (int)s;
	
		if (score_ >= Scores[3]) {
			rating_ = Ratings[3];
			rating_ += '!';
			return;
		}
		if (score_ >= Scores[2]) {
			rating_ = Ratings[2];
			return;
		}
		if (score_ > Scores[1]) {
			rating_ = Ratings[1];
			return;
		}
		rating_ = Ratings[0];
	}
}
